import asyncio
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from enum import Enum, auto
from typing import List, Any, Awaitable, Union, cast, Optional

from aflowey import AsyncFlow, aflow
from aflowey.single_executor import SingleFlowExecutor
from aflowey.types import AnyCallable, Executor

FlowOrListFlow = Union[List[AsyncFlow], AsyncFlow]


class ExecutorType(Enum):
    THREAD_POOL = auto()
    PROCESS_POOL = auto()


class AsyncFlowExecutor:
    """
    Execute several flows concurrently

    >>>await (aexec().from_flows(flows) | flow).run()

    """

    def __init__(
        self, /, executor: Union[Executor, Optional[ExecutorType]] = None, **kwargs: Any
    ) -> None:
        """
        Creates a new async flow executor
        Args:
            executor: the executor to process synchronous code, could be a ThreadPoolExecutor
            ProcessPoolExecutor
            kwargs: all passed to create the executor
        """
        self.executor = AsyncFlowExecutor._init_executor_if_needed(executor, **kwargs)
        self.flows: List[Union[AsyncFlow, List[AsyncFlow]]] = []

    def __or__(self, flow: List[AsyncFlow]) -> "AsyncFlowExecutor":
        """add a flow to execute in parallel"""
        self.flows.append(flow)

        return self

    def __enter__(self) -> "AsyncFlowExecutor":
        """if no executor provided, raise an error as the use of the with
        keyword is useless. Creates the context of the current executor"""
        if self.executor is None:
            raise ValueError(
                "Trying to use with context with not executor provided"
            )  # pragma: no cover
        self.executor.__enter__()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """closing the current executor"""
        if self.executor is None:
            raise ValueError(
                "Trying to use with context with not executor provided"
            )  # pragma: no cover
        self.executor.__exit__(exc_type, exc_val, exc_tb)

    @staticmethod
    async def _execute_one_flow(flow: AsyncFlow, executor: Executor) -> Any:
        """Run"""
        return await SingleFlowExecutor(flow, executor).execute_flow()

    def _execute_or_gather(self, flow: FlowOrListFlow) -> Awaitable[Any]:
        if isinstance(flow, list):
            flows_task = [self._execute_one_flow(flow, self.executor) for flow in flow]
            return asyncio.gather(*flows_task)
        return self._execute_one_flow(flow, self.executor)

    def run(self, **kwargs: Any) -> Any:
        """main function to run stuff in parallel"""
        # if other flow run in parallel
        flows = [self._execute_or_gather(flow) for flow in self.flows]
        return asyncio.gather(*flows, **kwargs)

    @staticmethod
    def _init_executor_if_needed(
        executor: Optional[Union[Executor, ExecutorType]], **kwargs: Any
    ) -> Optional[Executor]:
        if executor is None:
            return None
        if isinstance(executor, ExecutorType):
            if executor is ExecutorType.THREAD_POOL:
                return ThreadPoolExecutor(**kwargs)
            elif executor is ExecutorType.PROCESS_POOL:
                return ProcessPoolExecutor(**kwargs)
            raise ValueError("Wrong provided executor type")
        return executor

    @staticmethod
    def ensure_flow(fn: Any, arg: Any = None) -> AsyncFlow:
        if isinstance(fn, AsyncFlow):
            return fn
        flow = aflow.empty() if arg is None else aflow.from_args(arg)
        return cast(AsyncFlow, flow >> fn)

    def from_flows(self, flows: Any) -> "AsyncFlowExecutor":
        """create a new executor from one flow or array of flows"""
        if not isinstance(flows, list):
            flows = AsyncFlowExecutor.ensure_flow(flows)
            self.flows.append(flows)
        else:
            new_flows = [AsyncFlowExecutor.ensure_flow(value) for value in flows]
            self.flows.append(new_flows)
        return self

    @staticmethod
    def starmap(
        *flows: Any, executor: Optional["AsyncFlowExecutor"] = None
    ) -> AnyCallable:
        async def wrapper(arg: Any) -> Any:
            new_flows = [AsyncFlowExecutor.ensure_flow(fn, arg) for fn in flows]
            executor_inst = executor.executor if executor is not None else executor
            result = (
                await AsyncFlowExecutor(executor=executor_inst)
                .from_flows(new_flows)
                .run()
            )
            return result[0]

        return wrapper


async_exec = aexec = AsyncFlowExecutor
flows_from_arg = spawn_flows = run_flows = astarmap = AsyncFlowExecutor.starmap



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from typing import Any, TypeVar, Callable, Optional, Union, Coroutine


Opt = Optional
Function = TypeVar("Function", bound=Callable[..., Any])
Executor = Optional[Union[ThreadPoolExecutor, ProcessPoolExecutor]]
AnyCoroutineFunction = Callable[[Any, Any], Coroutine[Any, Any, Any]]
AnyCallable = Callable[[Any], Any]



import asyncio
import functools
import inspect
from typing import Any, Callable, List, Union, Iterable, Optional, cast, Type

from loguru import logger

from aflowey import F
from aflowey.types import Function


def async_wrap(func: Function) -> F:
    """wrap a function into a coroutine function
    Args:
        func: callable
    Returns:
        F instance
    """

    @functools.wraps(func)
    async def wrapper(*a: Any, **kw: Any) -> Any:
        return func(*a, **kw)

    return F(wrapper)


def log(log_str: str = "", print_arg: bool = False) -> Any:
    """utility function to log between steps, printing argument if needed"""

    async def wrapped(last_result: Any) -> Any:
        if log_str:
            logger.info(log_str)
        if print_arg:  # pragma: no cover
            logger.info(last_result)
        return last_result

    return wrapped


flog = log


def partial(func: Function, *args: Any, **kwargs: Any) -> F:
    """make a partial function of the given func and ensure
    it will work in an async context
    """
    return F(cast(Callable[[], Any], functools.partial(func, *args, **kwargs)))


p = apartial = partial


def f1(func: Function, extractor: Optional[Function] = None) -> F:
    """wraps a one argument function (with arity 1) and allows
    to add an extractor to work on the input argument.
    """

    is_coroutine = asyncio.iscoroutinefunction(func)

    @functools.wraps(func)
    def wrapped(arg1: Any) -> Any:
        value = arg1 if extractor is None else extractor(arg1)
        return func(value)

    wrapped_fn = async_wrap(wrapped) if is_coroutine else F(wrapped)
    wrapped_fn.__F1__ = True  # type: ignore

    return wrapped_fn


F1 = f1


def f0(func: Function) -> F:
    """create a new function from a 0 arity function (takes 0 args). The new
    function takes exactly one argument and does not pass it to the wrapped
    function. It allows using a 0 arity function in a flow relatively easily.
    """
    is_coroutine = asyncio.iscoroutinefunction(func)

    @functools.wraps(func)
    def wrapped(arg1: Any) -> Any:
        return func()

    wrapped_fn = async_wrap(wrapped) if is_coroutine else F(wrapped)
    wrapped_fn.__F0__ = True  # type: ignore

    return wrapped_fn


F0 = f0


def may_fail(func: Function) -> Function:
    """simply for readability"""
    return func


breaker = erratic = may_fail


def spread_args(func: Function) -> F:
    """create a function which takes an iterable of args
    and spread it into the given function"""

    is_coroutine = asyncio.iscoroutinefunction(func)

    @functools.wraps(func)
    def wrapped(args: Iterable[Any]) -> Any:
        # if to much args, slice the args to given length
        return func(*args)

    wrapped_fn = async_wrap(wrapped) if is_coroutine else F(wrapped)
    return wrapped_fn


spread = spread_args


def spread_kwargs(func: Function) -> F:
    """create a function which takes a mapping of kwargs
    and spread it into the given function"""

    is_coroutine = asyncio.iscoroutinefunction(func)

    @functools.wraps(func)
    def wrapped(**kwargs: Any) -> Any:
        arg_spec = inspect.getfullargspec(func)
        args = set(arg_spec.args + arg_spec.kwonlyargs) - set(["self"])
        new_kwargs = {key: kwargs[key] for key in args}
        return func(**new_kwargs)

    wrapped_fn = async_wrap(wrapped) if is_coroutine else F(wrapped)
    return wrapped_fn


spread_kw = spread_kwargs


def lift(f: Function, lift_op: Function = map) -> F:  # type: ignore[assignment]
    return p(lift_op, f)


def ensure_f(func: Function) -> F:
    """wrap the given function into a F instance"""
    if not isinstance(func, F):
        return F(func)
    return func


def make_impure(func: Union[Function, F]) -> F:
    """tags the given function as impure, i.e. consume an argument
    but does not create new one
    """
    # automatically create new function when the function
    # is a bound method and has no other input args
    func = ensure_f(func)
    func.__side_effect__ = True  # type: ignore
    return func


def side_effect(*func: Union[Function, F]) -> Union[List[F], F]:
    """
    take an array of function and tag it as side effects
    function
    """
    if len(func) == 1:
        return make_impure(func[0])
    return [make_impure(fu) for fu in func]


imp = impure = side_effect


def ensure_callable(x: Union[Any, Function]) -> Function:
    """ensure a given args is a callable by returning a new callable if not"""
    if not callable(x):

        def wrapped(*args: Any, **kwargs: Any) -> Any:
            return x

        return cast(Function, wrapped)
    return cast(Function, x)


def named(func: Union[Function, F], name: str) -> F:
    """tags a function as a named function"""
    if not callable(func):
        func = ensure_callable(func)
    func = ensure_f(func)
    func.__named__ = name  # type: ignore
    return func


def identity(x: Any) -> Any:
    return x


# helper functions
def is_f(func: Any) -> bool:
    return isinstance(func, F)


def is_f0(func: Function) -> bool:
    return hasattr(func, "__F0__")


def is_f1(func: Function) -> bool:
    return hasattr(func, "__F1__")


def is_side_effect(func: Function) -> bool:
    return hasattr(func, "__side_effect__")


def get_name(func: Function) -> str:
    if hasattr(func, "__named__"):
        return func.__named__  # type: ignore
    return ""



import asyncio
import functools
from typing import Any, Union, Awaitable

from loguru import logger

from aflowey import AsyncFlow, F
from aflowey.functions import is_f0, is_side_effect, get_name
from aflowey.types import Function, AnyCoroutineFunction, Executor


async def _exec(function: Union[F, Function], *a: Any, **kw: Any) -> Any:
    current_result = function(*a, **kw)
    while asyncio.iscoroutine(current_result):
        current_result = await current_result
    if callable(current_result):
        return await _exec(current_result)
    return current_result


def _exec_synchronously(fn: Function, *args: Any) -> Any:
    result = fn(*args)
    while callable(result):
        result = result()
    return result


def async_wrap(func: F) -> AnyCoroutineFunction:
    """wrap the given into a coroutine function and try
    calling it
    """

    @functools.wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        return await _exec(func, *args, **kwargs)

    return wrapped


def check_and_run_step(
    executor: Executor, fn: F, *args: Any, **kwargs: Any
) -> Awaitable[Any]:
    if fn.is_coroutine_function:
        new_fn = async_wrap(fn)
        return asyncio.create_task(new_fn(*args, **kwargs))  # type: ignore[call-arg]
    if executor is None:
        new_fn = async_wrap(fn)
        return new_fn(*args, **kwargs)  # type: ignore[call-arg]
    loop = asyncio.get_event_loop()
    new_fn = fn.func
    if kwargs:
        new_fn = functools.partial(new_fn, **kwargs)
    return loop.run_in_executor(executor, _exec_synchronously, new_fn, *args)


class SingleFlowExecutor:
    """Single flow executor"""

    CANCEL_FLOW = object()

    def __init__(self, flow: AsyncFlow, executor: Executor = None) -> None:
        self.flow = flow
        self.executor = executor

    @staticmethod
    async def check_and_execute_flow_if_needed(
        maybe_flow: Union[Any, AsyncFlow], executor: Executor
    ) -> Any:
        """check if we have an async flow and execute it"""
        if isinstance(maybe_flow, AsyncFlow):
            return await SingleFlowExecutor(maybe_flow, executor).execute_flow()
        return maybe_flow

    @staticmethod
    def need_to_cancel_flow(result: Any) -> bool:
        """check if we need to cancel flow checking sentinel"""
        if result is SingleFlowExecutor.CANCEL_FLOW:
            logger.debug("Received sentinel object, canceling flow...")
            return True
        return False

    @staticmethod
    def get_step_name(func: F, index: int) -> str:
        """get the step name"""
        return get_name(func) or str(index)

    def save_step(self, task: F, index: int, current_args: Any) -> None:
        """save step state in flow attribute"""
        self.flow.steps[self.get_step_name(task, index)] = current_args

    def _check_current_args_if_side_effect(self, first_aws: F, res: Any) -> Any:
        if is_side_effect(first_aws):
            return self.flow.kwargs if self.flow.kwargs else self.flow.args
        return res

    async def _execute_first_step(self, first_aws: F) -> Any:
        """executing the first step"""
        if not self.flow.args and not self.flow.kwargs and is_f0(first_aws):
            self.flow.args = (None,)

        res = await check_and_run_step(
            self.executor, first_aws, *self.flow.args, **self.flow.kwargs
        )
        current_args = self._check_current_args_if_side_effect(first_aws, res)
        # if flow run it
        current_args = await self.check_and_execute_flow_if_needed(
            current_args, self.executor
        )

        # memorize name
        self.save_step(first_aws, 0, current_args)
        return current_args

    async def execute_flow(self) -> Any:
        """Main function to execute a flow"""
        if not self.flow.aws:
            return None

        # get first step
        first_aws = self.flow.aws[0]
        current_args = await self._execute_first_step(first_aws)

        for index, task in enumerate(self.flow.aws[1:]):

            result = await check_and_run_step(self.executor, task, current_args)

            if is_side_effect(task):
                # side effect task, does not return a value
                self.save_step(task, index + 1, current_args)
                if self.need_to_cancel_flow(result):
                    break  # pragma: no cover
                continue  # pragma: no cover

            result = await self.check_and_execute_flow_if_needed(result, self.executor)
            if self.need_to_cancel_flow(result):  # check if we need to cancel the flow
                break

            current_args = result
            self.save_step(task, index + 1, current_args)
        # return current args that are the actual results
        return current_args


CANCEL_FLOW = SingleFlowExecutor.CANCEL_FLOW
